@@ -2,9 +2,9 @@
 
 const { test } = require('tap')
 const { BalancedPool, Pool, Client, errors } = require('..')
+const { nodeMajor } = require('../lib/core/util')
 const { createServer } = require('http')
 const { promisify } = require('util')
-const semver = require('semver')
 
 test('throws when factory is not a function', (t) => {
   t.plan(2)
@@ -437,7 +437,7 @@ const cases = [
     expectedRatios: [0.34, 0.34, 0.32],
 
     // Skip because the behavior of Node.js has changed
-    skip: semver.satisfies(process.version, '>= 19.0.0')
+    skip: nodeMajor >= 19
   },
 
   // 8
@@ -515,7 +515,16 @@ for (const [index, { config, expected, expectedRatios, iterations = 9, expectedC
       try {
         await client.request({ path: '/', method: 'GET' })
       } catch (e) {
-        const serverWithError = servers.find(server => server.port === e.port) || servers.find(server => server.port === e.socket.remotePort)
+        const serverWithError =
+          servers.find(server => server.port === e.port) ||
+          servers.find(server => {
+            if (typeof AggregateError === 'function' && e instanceof AggregateError) {
+              return e.errors.some(e => server.port === (e.socket?.remotePort ?? e.port))
+            }
+
+            return server.port === e.socket.remotePort
+          })
+
         serverWithError.requestsCount++
 
         if (e.code === 'ECONNREFUSED') {
